unit Hwang;

interface

{ $DEFINE DEBUG}

uses types,kreslic;

type tHwang=object(tPlayer)
                ActualTime: Longint;
                MemoryTime: array[1..32, 1..32] of Longint;
                MemoryData: array[1..32, 1..32] of TSquare;
                MemoryInfo: array[1..32, 1..32] of Integer;
                StrategMap: array[-1..1, -1..1, 1..8, 1..4] of Integer;
                DistanceMap: array[1..32, 1..32] of Integer;
                IntArray: array[1..3000] of Integer;
                TargetX, TargetY, TargetDirection: Integer;
                MoveDX, MoveDY, TurnDX, TurnDY: Integer;
                Weapon: Integer;
                Action: Integer;
                ExploreNearest: Boolean;
                CurDirection: Integer;
                constructor InitGame(MyNumber:integer; var name:string);
                procedure MakeTurn(var move_dx,move_dy,fire,turn_dx,turn_dy:shortint);virtual;
                function SquareBlocked(I, J: Integer): Boolean;
                procedure CalcDistanceMap;
                procedure CalcStrategicMap;
                procedure MapRange(var X1, Y1, X2, Y2: Integer);
                procedure UpdateMemory;
                procedure Explore;
                procedure MoveToTarget;
                procedure ChooseAction;
              end;
     pHwang=^THwang;

implementation

const
  actNone = 0;
  actExplore = 1;
  actGetItem = 2;
  actAttack = 3;

  DirectionTab: array[1..8, 1..2] of Shortint =
    ((-1,-1), (0,-1), (1,-1), (1,0), (1,1), (0,1), (-1,1), (-1,0));
  VisibleTab: array[1..8, 1..4] of Shortint = (
    (-3, -3, 0, 0), (-1, -4, 1, 0), (0, -3, 3, 0), (0, -1, 4, 1),
    (0, 0, 3, 3), (-1, 0, 1, 4), (-3, 0, 0, 3), (-4, -1, 0, 1));
  OffsetToDirection: array[-1..1, -1..1] of Shortint = (
    (1, 8, 7), (2, 0, 6), (3, 4, 5));

function IntToStr(I: Integer): String;
var
  S: String;
begin
  Str(I, S);
  IntToStr := S;
end;

function Min(I, J: Integer): Integer;
begin
  if I <= J then
    Min := I
  else
    Min := J;
end;

function Max(I, J: Integer): Integer;
begin
  if I >= J then
    Max := I
  else
    Max := J;
end;

procedure SayDebug(S: String);
begin
{$IFDEF DEBUG}
  Say(15, S);
{$ENDIF}
end;

constructor THwang.InitGame;
var
  I, J: Integer;
begin
  {FillChar(ActualTime, SizeOf(Self), 0);}
  Name:='Ufon #1';
  ActualTime := 0;
  for I := 1 to 32 do
    for J := 1 to 32 do
      MemoryTime[I, J] := -1000000;
  FillChar(MemoryData, SizeOf(MemoryData), 0);
  FillChar(DistanceMap, SizeOf(DistanceMap), 0);
  Action := actNone;
end;

function THwang.SquareBlocked(I, J: Integer): Boolean;
begin
  SquareBlocked :=
    (I < 1) or (J < 1) or (I > 32) or (J > 32) or Walls[I, J] or
    ((MemoryTime[I, J] = ActualTime - 1) and (MemoryData[I, J].Who = Who_Player));
end;

procedure THwang.CalcDistanceMap;
var
  I, J, K, L, M: Integer;
  FrBegin, FrEnd: Integer;
begin
  for I := 1 to 32 do
    for J := 1 to 32 do
      if SquareBlocked(I, J) then
        DistanceMap[I, J] := -2
      else
        DistanceMap[I, J] := -1;
  DistanceMap[X, Y] := 0;
  FrBegin := 1;
  FrEnd := 1;
  IntArray[1] := X + Y shl 8;
  while FrBegin <= FrEnd do
  begin
    I := IntArray[FrBegin];
    J := Hi(I);
    I := Lo(I);
    Inc(FrBegin);
    for K := 1 to 8 do
    begin
      L := I + DirectionTab[K, 1];
      M := J + DirectionTab[K, 2];
      if DistanceMap[L, M] = -1 then
      begin
        DistanceMap[L, M] := DistanceMap[I, J] + 1;
        Inc(FrEnd);
        IntArray[FrEnd] := L + M shl 8;
      end;
    end;
  end;
end;

procedure THwang.MapRange(var X1, Y1, X2, Y2: Integer);
begin
  if X1 < 1 then
    X1 := 1;
  if Y1 < 1 then
    Y1 := 1;
  if X2 > 32 then
    X2 := 32;
  if Y2 > 32 then
    Y2 := 32;
end;

procedure THwang.UpdateMemory;
var
  I, J, X1, Y1, X2, Y2: Integer;
begin
  I := OffsetToDirection[DX, DY];
  X1 := X + VisibleTab[I, 1];
  Y1 := Y + VisibleTab[I, 2];
  X2 := X + VisibleTab[I, 3];
  Y2 := Y + VisibleTab[I, 4];
  MapRange(X1, Y1, X2, Y2);
  for I := X1 to X2 do
    for J := Y1 to Y2 do
    begin
      MemoryTime[I, J] := ActualTime;
      MemoryData[I, J] := Things[I, J];
    end;
  MemoryData[X, Y].Who := 0;
  MemoryData[X, Y].What := 0;
end;

procedure THwang.CalcStrategicMap;
var
  I, J, K, L, Att, Def, Range, XX, YY: Integer;
begin
  for I := -1 to 1 do
    for J := -1 to 1 do
      for K := 1 to 8 do
      begin
        Att := 0;
        Def := 10;
        Range := 0;
        for L := 0 to 5 do
        begin
          if Walls[X + I + L * DirectionTab[K, 1],
                   Y + J + L * DirectionTab[K, 2]] then
            Break;
          Inc(Range);
        end;
        if not Walls[X + I, Y + J] and
           not ((MemoryTime[X + I, Y + J] = ActualTime) and
            (MemoryData[X + I, Y + J].Who = Who_Player)) then
        begin
        for L := 1 to 5 do
        begin
          XX := X + I + L * DirectionTab[K, 1];
          YY := Y + J + L * DirectionTab[K, 2];
          if Walls[XX, YY] then
            Break;
          if (MemoryTime[XX, YY] = ActualTime) and
            (MemoryData[XX, YY].Who = Who_Player) then
          begin
            Att := Max(Att, 6 - L);
            if K = OffsetToDirection[-Things[XX, YY].DX,
                                     -Things[XX, YY].DY] then
              Dec(Def);
          end;
        end; end;
        StrategMap[I, J, K, 1] := Att;
        StrategMap[I, J, K, 2] := Def;
        StrategMap[I, J, K, 3] := Range;
        StrategMap[I, J, K, 4] := 0;
      end;
end;

procedure THwang.Explore;
var
  I, J, L: Integer;
  K, M: Longint;
  D, E: Integer;
begin
  ExploreNearest := {Lives < 8}True;
  Action := actExplore;
  K := ActualTime;
  L := 0;
  D := 0;
  for I := 1 to 32 do
    for J := 1 to 32 do
    begin
      E := DistanceMap[I, J];
      if E > 0 then
      begin
        M := MemoryTime[I, J];
        if (M = K) and (not ExploreNearest or (E = D)) then
          Inc(L)
        else if (M < K) or ((M = K) and (E < D)) then
        begin
          K := M;
          L := 1;
          D := E;
        end;
      end;
    end;
  L := Random(L);
  for I := 1 to 32 do
    for J := 1 to 32 do
    begin
      E := DistanceMap[I, J];
      if (E > 0) and (not ExploreNearest or (E = D)) then
        if MemoryTime[I, J] = K then
        begin
          if L = 0 then
          begin
            TargetX := I;
            TargetY := J;
            Exit;
          end;
          Dec(L);
        end;
      end;
  {TargetX := X;
  TargetY := Y;}
end;

procedure THwang.MoveToTarget;
var
  I, J, K, L: Integer;
begin
  if DistanceMap[TargetX, TargetY] < 0 then
  begin
    MoveDX := 0;
    MoveDY := 0;
  end;
  I := TargetX;
  J := TargetY;
  TurnDX := -2;
  while DistanceMap[I, J] > 1 do
  begin
    L := DistanceMap[I, J] - 1;
    for K := 1 to 8 do
    begin
      if L = DistanceMap[I + DirectionTab[K, 1], J + DirectionTab[K, 2]] then
      begin
        TurnDX := - DirectionTab[K, 1];
        TurnDY := - DirectionTab[K, 2];
        Inc(I, DirectionTab[K, 1]);
        Inc(J, DirectionTab[K, 2]);
        Break;
      end;
    end;
  end;
  MoveDX := I - X;
  MoveDY := J - Y;
  if TurnDX = -2 then
  begin
    TurnDX := MoveDX;
    TurnDY := MoveDY;
  end;
end;

procedure THwang.ChooseAction;
var
  Tab: array[1..8] of Integer;
  I, J, K, L, M: Integer;
  ActDistance: Integer;
  BestI, BestJ, BestK, BestAtt, BestDef, BestDis: Integer;
  F, BestFire: Integer;
begin
  if Action <> actExplore then
    if (Action = actNone) or ((TargetX <> X) and (TargetY <> Y)) then
    Explore;
  {Action := actExplore;}
  ActDistance := 10000;
  for I := 1 to 32 do
    for J := 1 to 32 do
    begin
      K := MemoryData[I, J].What;
      L := MemoryData[I, J].Who;
      M := DistanceMap[I, J];
      if (L = Who_Player) then
      begin
        if (Lives >= 3) and (M < ActDistance) and (MemoryTime[I, J] >= ActualTime - 10) then
        begin
          Action := actAttack;
          ActDistance := DistanceMap[I, J];
          TargetX := I;
          TargetY := J;
        end
      end
      else if (K >= Uzi) and (K <= Detonator) and (
        (Action <> actAttack) or (M < ActDistance + 8)) and (M < ActDistance) then
      begin
        Action := actGetItem;
        ActDistance := DistanceMap[I, J];
        TargetX := I;
        TargetY := J;
      end
    end;
  if (TargetX = X) and (TargetY = Y) then
    Explore;
  if DistanceMap[TargetX, TargetY] < 0 then
    if Action <> actAttack then
      Explore;

  MoveToTarget;
  if (TurnDX < -1) or (TurnDY < -1) or (TurnDX > 1) or (TurnDY > 1) then
  begin TurnDX := 0; TurnDY := 0; end;
  if (MoveDX < -1) or (MoveDY < -1) or (MoveDX > 1) or (MoveDY > 1) then
  begin MoveDX := 0; MoveDY := 0; end;
  CurDirection := OffsetToDirection[TurnDX, TurnDY];


  Weapon := Gun;
  if (Action = actAttack) then
  begin
    BestAtt := StrategMap[MoveDX, MoveDY, CurDirection, 1];
    BestDef := StrategMap[MoveDX, MoveDY, CurDirection, 2];
    BestDis := 0;
    BestI := MoveDX;
    BestJ := MoveDY;
    BestK := 0;
    BestFire := BestAtt;
    for I := -1 to 1 do
      for J := -1 to 1 do
        for K := 1 to 8 do
        begin
          F := StrategMap[I, J, CurDirection, 1];
          L := StrategMap[I, J, K, 1];
          if (F > BestFire) or
             ((F = BestFire) and (L > BestAtt)) then
          {or
             ((L > 0) and (StrategMap[I, J, CurDirection, 2] > BestDef))}
          begin
            BestI := I;
            BestJ := J;
            {BestDis := DistanceMap[X + I, Y + J];}
            BestK := CurDirection;
            BestFire := F;
            BestAtt := L;
            BestDef := StrategMap[I, J, CurDirection, 2];
          end;
      end;
    SayDebug('A'+IntToStr(BestAtt)+'D'+IntToStr(BestDef)+'I'+IntToStr(BestI)+'J'+IntToStr(BestJ));
    if BestK <> 0 then
    begin
      MoveDX := BestI;
      MoveDY := BestJ;
      {TargetDirection := BestK;}
      for I := 1 to 8 do
        if StrategMap[BestI, BestJ, I, 1] > 0 then
        begin
          TargetDirection := I;
          Break;
        end;
    end;
  end
  else
  begin
    J := 0;
    TargetDirection := OffsetToDirection[TurnDX, TurnDY];
    for I := 1 to 8 do
    begin
      if I = TargetDirection then
        Inc(J, 15);
      Inc(J, StrategMap[X + MoveDX, Y + MoveDY, I, 3] + 2);
    end;
    J := Random(J);
    for I := 1 to 8 do
    begin
      if I = TargetDirection then
        Dec(J, 15);
      Dec(J, StrategMap[X + MoveDX, Y + MoveDY, I, 3] + 2);
      if J <= 0 then
      begin
        TargetDirection := 0;
        Break;
      end;
    end;
  end;

  if ((Rations[Uzi] > 0) and
     (StrategMap[MoveDX, MoveDY, CurDirection, 1] > 0))
     or ((Rations[Uzi] > 20) and (Action = actAttack)) then
    Weapon := Uzi
  else
    Weapon := Gun;
end;

procedure THwang.MakeTurn;
var
  OldAction: Integer;
  I: Integer;
begin

  IF TURNNUMBER mod 10=0 then say(random(8)+1);

  CurDirection := OffsetToDirection[DX, DY];
  TargetDirection := 0;
  Weapon := Gun;

  OldAction := Action;
  UpdateMemory;
  CalcDistanceMap;
  CalcStrategicMap;
  ChooseAction;

  {MoveToTarget;}

  Fire := Weapon;

  if (Action <> actAttack) and (Random(4) = 0) then
    TargetDirection := Random(8) + 1;
  if TargetDirection <> 0 then
  begin
    Turn_DX := DirectionTab[TargetDirection, 1];
    Turn_DY := DirectionTab[TargetDirection, 2];
  end
  else
  begin
    Turn_DX := TurnDX;
    Turn_DY := TurnDY;
  end;

  Move_DX := MoveDX;
  Move_DY := MoveDY;
  Died:=False;

  {if Action <> OldAction then
    case Action of
    actAttack: Say(15, 'Atacking');
    actGetItem: Say(15, 'Taking');
    actExplore: Say(15, 'Exploring');
    end;}
  OldAction := Action;

  Inc(ActualTime);
end;

begin
end.